/*
Copyright (C) 2006-2009 Noah Graham, all rights reserved.
See license.txt for permissible uses.

Refer to 

N. Graham, ``An Electroweak Oscillon,'' hep-th/0610267, Phys. Rev. Lett. 98 (2007) 101801
and/or
N. Graham, arXiv:0706.4125 [hep-th], ``Numerical Simulation of an Electroweak Oscillon,'' Phys. Rev.  D 76 (2007) 085017

in any publications or other products making use of this code.
*/

#ifndef __COMPLEX_H
#define __COMPLEX_H

#include "style.h"
#include <math.h>

class Complex
{
public:
  Complex() : real(0.0), imag(0.0) {}
  Complex(FloatType _real) : real(_real), imag(0.0) {}
  Complex(FloatType _real, FloatType _imag) : real(_real), imag(_imag)  {}
  Complex(const Complex &c): real(c.real), imag(c.imag)  {}

  Complex conjugate() const
  {
    return Complex(real, -imag);
  }

  const Complex &operator++()
  {
    real++;
    return *this;
  }

  const Complex &operator--()
  {
    real--;
    return *this;
  }


  Complex operator+(FloatType b) const
  {
    return Complex(real + b, imag);
  }

  Complex operator+(Complex c) const
  {
    return Complex(real + c.real, imag + c.imag);
  }

  const Complex &operator+=(Complex c)
  {
    return ((*this) = (*this) + c);
  }

  const Complex &operator+=(FloatType d)
  {
    return ((*this) = (*this) + d);
  }

  Complex operator-(FloatType d) const
  {
    return Complex(real - d, imag);
  }

  Complex operator-(Complex c) const
  {
    return Complex(real - c.real, imag - c.imag);
  }

  const Complex &operator-=(FloatType d)
  {
    return ((*this) = (*this) - d);
  }

  const Complex &operator-=(Complex c)
  {
    return ((*this) = (*this) - c);
  }

  Complex operator*(FloatType d) const
  {
    return Complex(real*d, imag*d);
  }

  Complex operator*(Complex c) const
  {
    return Complex(real * c.real - imag * c.imag,
                   imag * c.real + real * c.imag);
  }

  const Complex &operator*=(FloatType d)
  {
    return ((*this) = (*this)*d);
  }

  const Complex &operator*=(Complex c)
  {
    return ((*this)=(*this)*c);
  }

  Complex operator/(FloatType d) const
  {
    return Complex(real/d, imag/d);
  }

  Complex operator/(Complex denom) const
  {
    FloatType newreal, newimag;
    if (fabs(denom.real) >= fabs(denom.imag))
    {
        FloatType r = denom.imag/denom.real;
        FloatType fact = (denom.real + r * denom.imag);
        newreal = (real + r*imag)/fact;
        newimag = (imag - r*real)/fact;
    } else {
        FloatType r = denom.real/denom.imag;
        FloatType fact = (denom.imag + r * denom.real);
        newreal = (r * real + imag)/fact;
        newimag = (r * imag - real)/fact;
    }
    return Complex(newreal, newimag);
  }

  const Complex &operator/=(FloatType d)
  {
    return ((*this) = (*this)/d);
  }

  const Complex &operator/=(Complex denom)
  {
    return ((*this) = (*this)/denom);
  }

  Complex operator-() const { return Complex(-real, -imag); }
  Complex operator+() const { return *this; }

  static const Complex i;
  FloatType real_part() const { return real; }
  FloatType imag_part() const { return imag; }
  void set_real_part(FloatType d) { real = d; }
  void set_imag_part(FloatType d) { imag = d; }
  FloatType norm_squared() const { return real*real + imag*imag; }
  FloatType norm() const
  {
    FloatType areal = fabs(real);
    FloatType aimag = fabs(imag);
    if (areal >= aimag)
    {
        if (areal == 0.0) { return 0.0; }
        return areal * sqrt(1 + (imag*imag)/(real*real));
    } else {
        if (aimag == 0.0) { return 0.0; }
        return aimag * sqrt(1 + (real*real)/(imag*imag));
    }
  }
  FloatType abs_norm() const { return fabs(real) + fabs(imag); }
  FloatType arg() const { return atan2(imag, real); }

 protected:
    FloatType real;
    FloatType imag;
};

ostream &operator<<(ostream &o, Complex c)
{
  o << c.real_part() << "+" << c.imag_part() <<"*i";
  return o;
}


inline Complex operator+(FloatType d, Complex c) { return c + d; }

inline Complex operator-(FloatType d, Complex c) { return -c + d; }

inline Complex operator*(FloatType d, Complex c) { return c*d; }

inline Complex operator/(FloatType d, Complex c) { return Complex(d)/c; }


inline Complex log(Complex c)
{
  return log(c.norm()) + Complex::i*c.arg();
}

inline Complex exp(Complex c)
{
  FloatType a = c.imag_part();
  return exp(c.real_part())*(cos(a) + Complex::i*sin(a));
}

#endif
