/*
Copyright (C) 2006-2009 Noah Graham, all rights reserved.
See license.txt for permissible uses.

Refer to 

N. Graham, ``An Electroweak Oscillon,'' hep-th/0610267, Phys. Rev. Lett. 98 (2007) 101801
and/or
N. Graham, arXiv:0706.4125 [hep-th], ``Numerical Simulation of an Electroweak Oscillon,'' Phys. Rev.  D 76 (2007) 085017

in any publications or other products making use of this code.
*/

// #define DEBUG 1
#define DUMP_DATA 1

//Whether to use logarithm or linearized formula
#define LINEARIZE 1

#if NUM_PROCESSES != 1
#include "mpi.h"
#endif

#if NUM_THREADS != 1
#include "pthread.h"
#endif

typedef unsigned int uint;
typedef double FloatType;

#include "complex.h"

const Complex Complex::i(FloatType(0.0), FloatType(1.0));

// Direction in space
enum Direction {x, y, z};
// Direction in isospin space
enum SigmaDirection {sx, sy, sz};
// Indexes Higgs components
enum UpDown {up, down};

// Lattice spacing
const FloatType delta = FloatType(1);
// Time step
const FloatType h = FloatType(0.05);
// Lattice size
const FloatType xmax = FloatType(32.0);
// Dimension of lattice
const uint N = uint(xmax/delta + FloatType(0.5));
// Slice used in output routines
const uint DUMP_SLICE = N/2;
// SU(2) coupling constant
const FloatType g = sqrt(FloatType(2.0));
// U(1) coupling constant
const FloatType gp = 0.773;
// Higgs vev
const FloatType v = FloatType(1.0)/sqrt(FloatType(2.0));
// Higgs self-coupling
const FloatType lambda = FloatType(1.0);
// Factors used in Wilson line
const FloatType wilsonFact = g*delta/FloatType(2.0);
const FloatType wilsonFactU1 = gp*delta/FloatType(2.0);
// Box size used to monitor oscillon formation
const FloatType rbox = FloatType(28.0);
// Scaling of initial conditions
const FloatType eps = FloatType(1.15);

// Utilities for managing MPI communications
#if NUM_PROCESSES != 1
const int tag_spacer = 512;
const int tag_sign_spacer = 2048;
const int tally_tag_spacer = 4096;
#define CONTROL_RANK MpiRank(0)
class MpiRank
{
 public:
    explicit MpiRank(int rank) : mRank(rank) {}
    operator int() const { return mRank; }
    MpiRank next() const { return MpiRank((mRank + 1) % NUM_PROCESSES); }
    MpiRank prev() const { return MpiRank((mRank + NUM_PROCESSES - 1) % NUM_PROCESSES); }
 protected:
    int mRank;
};
MpiRank my_rank(0);
#else
#define MpiRank int
#define my_rank 0
#endif

// defines which slices we are responsible for computing
inline uint startpoint(MpiRank rank) { return (rank*N)/NUM_PROCESSES; }
inline uint endpoint(MpiRank rank)  { return ((rank+1)*N)/NUM_PROCESSES; }
// 2 extra to store the data passed from our neighbors
inline uint slice_size(MpiRank rank) { return endpoint(rank) - startpoint(rank) + 2; }

inline bool hasDumpSlice(MpiRank rank) { return (DUMP_SLICE >= startpoint(rank) && DUMP_SLICE < endpoint(rank)); }


class SiteLattice;
class LinkLattice;

// Structure for passing arguments to thread functions and associated bookkeeping
struct ArgStruct
{
    uint threadnum;
    const SiteLattice *phi0;
    const LinkLattice *A0;
    const SiteLattice *phi1;
    const LinkLattice *A1;
    SiteLattice *phi2;
    LinkLattice *A2;
#if NUM_PROCESSES != 1 && NUM_THREADS != 1
    static uint leftDoneCount;
    static uint rightDoneCount;
    static pthread_mutex_t synchMutex;
#endif
};

#if NUM_PROCESSES != 1 && NUM_THREADS != 1
uint ArgStruct::leftDoneCount;
uint ArgStruct::rightDoneCount;
pthread_mutex_t ArgStruct::synchMutex;
#endif

void setupArgs(ArgStruct *args,
               const SiteLattice &phi0,
               const SiteLattice &phi1, SiteLattice &phi2,
               const LinkLattice &A0, const LinkLattice &A1,
               LinkLattice &A2)
{
    for (uint n=0; n < NUM_THREADS; n++)
    {
        args[n].threadnum = n;
        args[n].phi0 = &phi0;
        args[n].A0 = &A0;
        args[n].phi1 = &phi1;
        args[n].A1 = &A1;
        args[n].phi2 = &phi2;
        args[n].A2 = &A2;
    }
}


// Fundamental Higgs
class Su2Fundamental
{
 public:
    Su2Fundamental(const Complex &upval, const Complex &downval)
    {
        mFund[up] = upval;
        mFund[down] = downval;
    }
    Su2Fundamental() {}

    Complex &operator[](UpDown i) {
        return mFund[i];
    }
    const Complex &operator[](UpDown i) const {
        return mFund[i];
    }
    Complex operator*(const Su2Fundamental &p2) const
    {
        return mFund[up].conjugate()*p2[up] + mFund[down].conjugate()*p2[down];
    }

    Su2Fundamental operator*(const Complex &c) const
    {
        return Su2Fundamental(mFund[up]*c, mFund[down]*c);
    }

    Su2Fundamental operator/(const Complex &c) const
    {
        return Su2Fundamental(mFund[up]/c, mFund[down]/c);
    }

    Su2Fundamental operator+(const Su2Fundamental &p2) const
    {
        return Su2Fundamental(mFund[up] + p2[up], mFund[down] + p2[down]);
    }

    const Su2Fundamental &operator+=(const Su2Fundamental &p2)
    {
        return (*this) = (*this) + p2;
    }

    Su2Fundamental operator-(const Su2Fundamental &p2) const
    {
        return Su2Fundamental(mFund[up] - p2[up], mFund[down] - p2[down]);
    }

    const Su2Fundamental &operator-=(const Su2Fundamental &p2)
    {
        return (*this) = (*this) - p2;
    }

    Su2Fundamental operator*(const FloatType f) const
    {
        return Su2Fundamental(mFund[up]*f, mFund[down]*f);
    }

    const Su2Fundamental &operator*=(const FloatType f)
    {
        return (*this) = (*this)*f;
    }

    Su2Fundamental operator/(const FloatType f) const
    {
        return Su2Fundamental(mFund[up]/f, mFund[down]/f);
    }

    const Su2Fundamental &operator/=(const FloatType f)
    {
        return (*this) = (*this)/f;
    }

    FloatType normSquared() const
    {
        return mFund[up].norm_squared() + mFund[down].norm_squared();
    }

    FloatType norm() const
    {
        return sqrt(normSquared());
    }

 protected:
    Complex mFund[2];
};

#ifdef DUMP_DATA
ostream &operator<<(ostream &os, const Su2Fundamental &phi)
{
    os << phi[up].real_part() << " "
       << phi[up].imag_part() << " "
       << phi[down].real_part() << " "
       << phi[down].imag_part();
    return os;
}
#endif
    
inline
Su2Fundamental operator*(FloatType f, const Su2Fundamental &phi)
{
    return phi*f;
}

inline
Su2Fundamental operator*(const Complex &c, const Su2Fundamental &phi)
{
    return phi*c;
}

// U(1) gauge field, algebra representation
class U1Linear
{
 public:
    U1Linear() : mVal(FloatType(0.0)) {}
    U1Linear(FloatType xval) : mVal(xval) {}

    U1Linear conjugate() const
    {
        return U1Linear(-mVal);
    }
    U1Linear operator-() const
    {
        return conjugate();
    }

    U1Linear operator+(const U1Linear &lin2) const
    {
        return U1Linear(mVal + lin2.mVal);
    }

    U1Linear &operator+=(const U1Linear &lin2)
    {
        mVal += lin2.mVal;
        return *this;
    }

    U1Linear operator-(const U1Linear &lin2) const
    {
        return U1Linear(mVal - lin2.mVal);
    }

    U1Linear &operator-=(const U1Linear &lin2)
    {
        mVal -= lin2.mVal;
        return *this;
    }

    U1Linear operator*(FloatType f) const
    {
        return U1Linear(mVal*f);
    }

    U1Linear &operator*=(FloatType f)
    {
        mVal *= f;
        return *this;
    }

    U1Linear operator/(FloatType f) const
    {
        return U1Linear(mVal/f);
    }

    U1Linear &operator/=(FloatType f)
    {
        mVal /= f;
        return *this;
    }

    FloatType len() const
    {
        return fabs(mVal);
    }

    FloatType lenSquared() const
    {
        return mVal*mVal;
    }

    FloatType val() const { return mVal; }

 protected:
    FloatType mVal;
    
};

inline
U1Linear operator*(FloatType f, const U1Linear &u1)
{
    return u1*f;
}


// U(1) gauge field, group representation
class U1
{

 public:
    U1() : mScaledVal(FloatType(0.0)) {}
    U1(U1Linear u1l) : mScaledVal(u1l.val()*wilsonFactU1) {}

    U1 conjugate() const
    {
        return U1(-tolinear());
    }

    U1 operator-() const
    {
        return conjugate();
    }

    U1 operator*(const U1 &other) const
    {
        return U1(tolinear() + other.tolinear());
    }

    U1 &operator*=(const U1 &other)
    {
        (*this) = (*this)*other;
        return *this;
    }

    // Addition (in algebra) and multiplication (in group) are the same in
    // the abelian theory.  Subtraction is the same as division is the same
    // as multiplication by the conjugate.
    U1 operator+(const U1 &other) const
    {
        return (*this)*other;
    }

    U1 &operator+=(const U1 &other)
    {
        (*this) *= other;
        return *this;
    }

    U1 operator/(const U1 &other) const
    {
        return (*this)*(other.conjugate());
    }

    U1 &operator/=(const U1 &other)
    {
        (*this) *= other.conjugate();
        return *this;
    }

    U1 operator-(const U1 &other) const
    {
        return (*this)*other.conjugate();
    }

    U1 &operator-=(const U1 &other)
    {
        (*this) *= other.conjugate();
        return *this;
    }

    U1 operator*(FloatType f) const
    {
        return U1(tolinear()*f);
    }

    U1 operator/(FloatType f) const
    {
        return U1(tolinear()/f);
    }

    Su2Fundamental operator*(const Su2Fundamental &phi) const
    {
        Complex u1 = cos(mScaledVal) + Complex::i*sin(mScaledVal);
        return Su2Fundamental(u1*phi[up], u1*phi[down]);
    }

    FloatType val() const { return mScaledVal/wilsonFactU1; }
    U1Linear tolinear() const {
        return U1Linear(val());
    }

    FloatType len() const
    {
        return fabs(val());
    }

    FloatType lenSquared() const
    {
        FloatType length = val();
        return length*length;
    }

 protected:
    FloatType mScaledVal;
};

#ifdef DUMP_DATA
ostream &operator<<(ostream &os, const U1 &u1)
{
    os << u1.val();
    return os;
}
#endif


inline
U1 operator*(FloatType f, const U1 &u1)
{
    return u1*f;
}

// SU(2) gauge field, algebra representation
class Su2AdjointLinear
{
 public:
    Su2AdjointLinear() { mAdj[sx] = mAdj[sy] = mAdj[sz] = FloatType(0.0); }
    Su2AdjointLinear(FloatType xval, FloatType yval, FloatType zval)
    {
        mAdj[sx] = xval;
        mAdj[sy] = yval;
        mAdj[sz] = zval;
    }

    FloatType &operator[](SigmaDirection i) {
        return mAdj[i];
    }
    const FloatType &operator[](SigmaDirection i) const {
        return mAdj[i];
    }

    Su2AdjointLinear conjugate() const
    {
        return Su2AdjointLinear(-mAdj[sx],
                                -mAdj[sy],
                                -mAdj[sz]);
    }
    Su2AdjointLinear operator-() const
    {
        return conjugate();
    }

    Su2AdjointLinear operator+(const Su2AdjointLinear &adj2) const
    {
        const Su2AdjointLinear &adj1 = *this;
        return Su2AdjointLinear(adj1[sx] + adj2[sx],
                                adj1[sy] + adj2[sy],
                                adj1[sz] + adj2[sz]);
    }

    const Su2AdjointLinear &operator+=(const Su2AdjointLinear &adj2)
    {
        return (*this) = (*this) + adj2;
    }

    Su2AdjointLinear operator-(const Su2AdjointLinear &adj2) const
    {
        const Su2AdjointLinear &adj1 = *this;
        return Su2AdjointLinear(adj1[sx] - adj2[sx],
                                adj1[sy] - adj2[sy],
                                adj1[sz] - adj2[sz]);
    }

    const Su2AdjointLinear &operator-=(const Su2AdjointLinear &adj2)
    {
        return (*this) = (*this) - adj2;
    }

    Su2AdjointLinear operator*(FloatType f) const
    {
        const Su2AdjointLinear &adj1 = *this;
        return Su2AdjointLinear(adj1[sx]*f,
                                adj1[sy]*f,
                                adj1[sz]*f);
    }

    const Su2AdjointLinear &operator*=(FloatType f)
    {
        return (*this) = (*this)*f;
    }

    Su2AdjointLinear operator/(FloatType f) const
    {
        const Su2AdjointLinear &adj1 = *this;
        return Su2AdjointLinear(adj1[sx]/f,
                                adj1[sy]/f,
                                adj1[sz]/f);
    }

    const Su2AdjointLinear &operator/=(FloatType f)
    {
        return (*this) = (*this)/f;
    }

    FloatType lenSquared() const
    {
        return mAdj[sx]*mAdj[sx] + mAdj[sy]*mAdj[sy] + mAdj[sz]*mAdj[sz];
    }
    FloatType len() const
    {
        return sqrt(lenSquared());
    }
 protected:
    FloatType mAdj[3];
    
};

inline
Su2AdjointLinear operator*(FloatType f, const Su2AdjointLinear &adj)
{
    return adj*f;
}

// SU(2) gauge field, group representation
class Su2Adjoint
{
 protected:

    Su2Adjoint(const Complex &c1, const Complex &c2) : x1(c1), x2(c2) {}

    FloatType rawlenSq() const {
        return (x2.imag_part()*x2.imag_part()+
                x2.real_part()*x2.real_part()+
                x1.imag_part()*x1.imag_part());
    }

    FloatType rawlen() const {
        return sqrt(rawlenSq());
    }

#ifndef LINEARIZE
    FloatType conversionFactor() const
    {
        FloatType norm = rawlen();

        if (wilsonFact*norm == FloatType(0.0))
        {
            return FloatType(0.0);
        }
        return asin(norm)/(wilsonFact*norm);
    }
#else
    FloatType conversionFactor() const
    {
      return FloatType(1.0)/wilsonFact;
    }
#endif

    void initFromLinear(FloatType xval, FloatType yval, FloatType zval)
    {
#ifndef LINEARIZE
        FloatType theta = sqrt(xval*xval + yval*yval + zval*zval);
        if (theta == FloatType(0.0)) {
            x1 = FloatType(1.0);
        } else {
            FloatType costheta = cos(theta*wilsonFact);
            FloatType sinthetanorm = sin(theta*wilsonFact)/theta;
            x1 = Complex(costheta, sinthetanorm*zval);
            x2 = Complex(sinthetanorm*yval, sinthetanorm*xval);
        }
#else
        FloatType normsq = sqr(wilsonFact*xval) + sqr(wilsonFact*yval) + sqr(wilsonFact*zval);
	x1 = Complex(sqrt(FloatType(1.0) - normsq), wilsonFact*zval);
	x2 = Complex(wilsonFact*yval, wilsonFact*xval);
#endif
    }

 public:
    Su2Adjoint() :  x1(FloatType(1.0)) {}

    Su2Adjoint(FloatType xval, FloatType yval, FloatType zval)
    {
        initFromLinear(xval, yval, zval);
    }

    Su2Adjoint(const Su2AdjointLinear &adj)
    {
        initFromLinear(adj[sx], adj[sy], adj[sz]);
    }

    FloatType operator[](SigmaDirection i) const
    {
        FloatType factor = conversionFactor();

        switch(i)
        {
         case sx:
            return x2.imag_part()*factor;
         case sy:
            return x2.real_part()*factor;
         case sz:
            return x1.imag_part()*factor;
         default:
            weak_assert(0);
            return FloatType(0.0);
        }
    }

    Su2Adjoint conjugate() const
    {
        return Su2Adjoint(x1.conjugate(), -x2);
    }

    Su2AdjointLinear tolinear() const
    {
        FloatType factor = conversionFactor();

        return Su2AdjointLinear(x2.imag_part()*factor,
                                x2.real_part()*factor,
                                x1.imag_part()*factor);
    }

    Su2Adjoint operator+(const Su2Adjoint &adj2) const
    {
        return Su2Adjoint(tolinear() + adj2.tolinear());
    }

    Su2Adjoint operator-(const Su2Adjoint &adj2) const
    {
        return Su2Adjoint(tolinear() - adj2.tolinear());
    }


    Su2Adjoint operator*(FloatType f) const
    {
        return Su2Adjoint(tolinear()*f);
    }

    Su2Adjoint operator/(FloatType f) const
    {
        return Su2Adjoint(tolinear()/f);
    }

    // matrix multiplication on links
    Su2Adjoint operator*(const Su2Adjoint &a2) const;

    Su2Fundamental operator*(const Su2Fundamental &p) const;

    FloatType len() const
    {
        return rawlen()/wilsonFact;
    }

    FloatType lenSquared() const {
        FloatType length = len();
        return length*length;
    }

 protected:
    Complex x1;
    Complex x2;
};

#ifdef DUMP_DATA
ostream &operator<<(ostream &os, const Su2Adjoint &su2)
{
    os << su2[sx] << " " << su2[sy] << " " << su2[sz];
    return os;
}
#endif

inline
Su2Adjoint operator*(FloatType f, const Su2Adjoint &adj)
{
    return adj*f;
}

// matrix multiplication of links
inline
Su2Adjoint Su2Adjoint::operator*(const Su2Adjoint &a2) const
{
    return Su2Adjoint(x1*a2.x1 - x2*(a2.x2.conjugate()),
                      x1*a2.x2 + x2*(a2.x1.conjugate()));
}

// multiply Higgs by link matrix
inline
Su2Fundamental Su2Adjoint::operator*(const Su2Fundamental &p) const
{
    return Su2Fundamental(x1*p[up] + x2*p[down],
                          x1.conjugate()*p[down] - x2.conjugate()*p[up]);
}

// Sites contain just the Higgs
class Site : public Su2Fundamental
{
 public:
    Site() : Su2Fundamental() {}
    Site(const Su2Fundamental &su2f) : Su2Fundamental(su2f) {}
};

// Links contain SU(2) and U(1)
class LinkLinear
{
 public:
    LinkLinear() {}
    LinkLinear(const Su2AdjointLinear &su2a, const U1Linear &u1) : su2link(su2a), u1link(u1) {}
    LinkLinear conjugate() const {
        return LinkLinear(su2link.conjugate(), u1link.conjugate());
    }

    LinkLinear operator+(const LinkLinear &other) const
    {
        return LinkLinear(su2link + other.su2link, u1link + other.u1link);
    }

    const LinkLinear &operator+=(const LinkLinear &other)
    {
        return *this = *this + other;
    }

    LinkLinear operator-(const LinkLinear &other) const
    {
        return LinkLinear(su2link - other.su2link, u1link - other.u1link);
    }

    const LinkLinear &operator-=(const LinkLinear &other)
    {
        return *this = *this - other;
    }

    LinkLinear operator*(FloatType f) const
    {
        return LinkLinear(su2link*f, u1link*f);
    }

    LinkLinear operator/(FloatType f) const
    {
        return LinkLinear(su2link/f, u1link/f);
    }

    FloatType lenSquared() const {
        return su2link.lenSquared() + u1link.lenSquared();
    }
    FloatType len() const {
        return sqrt(lenSquared());
    }

 protected:
    friend class Link;
    Su2AdjointLinear su2link;
    U1Linear u1link;
};


inline
LinkLinear operator*(FloatType f, const LinkLinear &l)
{
    return l*f;
}

// Links contain SU(2) and U(1)
class Link
{
 public:
    Link() {}
    Link(const Su2Adjoint &su2a, const U1 &u1) : su2link(su2a), u1link(u1) {}
    Link(const LinkLinear &ll) : su2link(ll.su2link), u1link(ll.u1link) {}
    Link conjugate() const {
        return Link(su2link.conjugate(), u1link.conjugate());
    }

    Link operator+(const Link &other) const
    {
        return Link(su2link + other.su2link, u1link + other.u1link);
    }

    Link operator*(const Link &other) const
    {
        return Link(su2link * other.su2link, u1link * other.u1link);
    }

    Link operator-(const Link &other) const
    {
        return Link(su2link - other.su2link, u1link - other.u1link);
    }

    Link operator*(FloatType f) const
    {
        return Link(su2link*f, u1link*f);
    }

    Link operator/(FloatType f) const
    {
        return Link(su2link/f, u1link/f);
    }

    Su2Fundamental operator*(const Su2Fundamental &phi) const
    {
        return u1link*(su2link*phi);
    }

    FloatType lenSquared() const {
        return su2link.lenSquared() + u1link.lenSquared();
    }

    FloatType len() const {
        return sqrt(lenSquared());
    }

    LinkLinear tolinear() const
    {
        return LinkLinear(su2link.tolinear(), u1link.tolinear());
    }

#ifdef DUMP_DATA
    const Su2Adjoint & getSu2() const { return su2link; }
    const U1 & getU1() const { return u1link; }
#endif

 protected:
    Su2Adjoint su2link;
    U1 u1link;
};

#ifdef DUMP_DATA
ostream &operator<<(ostream &os, const Link &link)
{
    os << link.getSu2() << " " << link.getU1();
    return os;
}
#endif

inline
Link operator*(FloatType f, const Link &link)
{
    return link*f;
}

// Indexes into SiteLattice
class SiteIndex
{
 public:

    SiteIndex(uint i, uint j, uint k)
    {
        weak_assert(i < N);
        weak_assert(j < N);
        weak_assert(k < N);

        mIndices[x] = i;
        mIndices[y] = j;
        mIndices[z] = k;
    }
    SiteIndex(const SiteIndex &other)
    {
        mIndices[x] = other.mIndices[x];
        mIndices[y] = other.mIndices[y];
        mIndices[z] = other.mIndices[z];
    }

    uint getSlice() const { return mIndices[x]; }
    uint getOffset() const { return N*mIndices[y] + mIndices[z]; }


    // lattice arithmetic with periodic boundary conditions
    SiteIndex operator+(Direction d) const
    {
        SiteIndex ret = *this;
        ret.mIndices[d] = (mIndices[d] + 1) % N;
        return ret;
    }

    SiteIndex operator-(Direction d) const
    {
        SiteIndex ret = *this;
        ret.mIndices[d] = (mIndices[d] + N - 1) % N;
        return ret;
    }

 protected:
    uint mIndices[3];
};

// Indexes into LinkLattice
class LinkIndex : public SiteIndex
{
 public:
    LinkIndex(const SiteIndex &si, Direction dir) : SiteIndex(si), mDirection(dir) {}
    Direction getDirection() const { return mDirection; }

 protected:
    Direction mDirection;
};

inline uint getSliceIndex(uint xindex)
{
    return (xindex + N + 1 - startpoint(my_rank)) % N;
}

// Stores 3-d grid of links 
// Links are labelled by the endpoint closest to the origin and the
// direction in which it emanates from that point
class LinkLattice
{
 protected:
    Link *getSlice(uint xindex, Direction d)
    {
        return mpLinks + N*N*(d*slice_size(my_rank) + getSliceIndex(xindex));
    }
    const Link *getSlice(uint xindex, Direction d) const
    {
        return mpLinks + N*N*(d*slice_size(my_rank) + getSliceIndex(xindex));
    }

    void alloc()
    {
        mpLinks = new Link[slice_size(my_rank)*N*N*3];
    }

    void free()
    {
        delete [] mpLinks;
        mpLinks = NULL;
    }

 public:
    LinkLattice()
    {
        alloc();
    }
    ~LinkLattice() { free(); }

    Link &operator()(const LinkIndex &index) {
        return getSlice(index.getSlice(), index.getDirection())[index.getOffset()];
    }

    const Link &operator()(const LinkIndex &index) const {
        return getSlice(index.getSlice(), index.getDirection())[index.getOffset()];
    }

    const Link &operator()(const SiteIndex &index, Direction dir) const {
        return getSlice(index.getSlice(), dir)[index.getOffset()];
    }

    LinkLattice(const LinkLattice &other)
    {
        alloc();
        memcpy(mpLinks, other.mpLinks, 3*slice_size(my_rank)*N*N*sizeof(Link));
    }

    LinkLattice &operator=(const LinkLattice &other)
    {
        free();
        alloc();
        memcpy(mpLinks, other.mpLinks, 3*slice_size(my_rank)*N*N*sizeof(Link));
        return *this;
    }
#if NUM_PROCESSES != 1
    void send(MpiRank rankto)
    {
        for (Direction d = x; d <= z; d = Direction(d+1))
        {
            if (rankto == my_rank.next()) {
                MPI_Isend(getSlice((endpoint(my_rank)-1),d), N*N*sizeof(Link), MPI_BYTE, rankto,
			  my_rank + (d+1)*tag_spacer, MPI_COMM_WORLD, &mRightSend[d]);
            }
            if (rankto.next() == my_rank) {
                MPI_Isend(getSlice(startpoint(my_rank),d), N*N*sizeof(Link), MPI_BYTE, rankto,
                          my_rank + (d+1)*tag_spacer + tag_sign_spacer, MPI_COMM_WORLD, &mLeftSend[d]);
            }
	}
    }
    void receive(MpiRank rankfrom)
    {
        for (Direction d = x; d <= z; d = Direction(d+1))
        {
            if (rankfrom.next() == my_rank) {
                MPI_Irecv(getSlice(endpoint(rankfrom)-1,d), N*N*sizeof(Link), MPI_BYTE, rankfrom,
			 rankfrom + (d+1)*tag_spacer, MPI_COMM_WORLD, &mRightReceive[d]);
            }
            if (rankfrom == my_rank.next()) {
                MPI_Irecv(getSlice(startpoint(rankfrom),d), N*N*sizeof(Link), MPI_BYTE, rankfrom,
			 rankfrom + (d+1)*tag_spacer + tag_sign_spacer, MPI_COMM_WORLD, &mLeftReceive[d]);
            }
        }
    }

    void wait()
    {
        MPI_Waitall(3,mLeftSend,MPI_STATUS_IGNORE);
        MPI_Waitall(3,mLeftReceive,MPI_STATUS_IGNORE);
        MPI_Waitall(3,mRightReceive,MPI_STATUS_IGNORE);
        MPI_Waitall(3,mRightSend,MPI_STATUS_IGNORE);
    }

#endif

    static void rotateData(LinkLattice &ll1, LinkLattice &ll2, LinkLattice &ll3)
    {
        Link *temp = ll1.mpLinks;
        ll1.mpLinks = ll2.mpLinks;
        ll2.mpLinks = ll3.mpLinks;
        ll3.mpLinks = temp;
    }

 protected:
    Link *mpLinks;
#if NUM_PROCESSES != 1
    MPI_Request mLeftSend[3], mRightSend[3], mLeftReceive[3], mRightReceive[3];
#endif
};

// Stores 3-d grid of Sites
class SiteLattice
{
 public:
    SiteLattice()
    {
        alloc();
    }
    ~SiteLattice() { free(); }

    Site &operator()(const SiteIndex &index) {
        return getSlice(index.getSlice())[index.getOffset()];
    }

    const Site &operator()(const SiteIndex &index) const {
        return getSlice(index.getSlice())[index.getOffset()];
    }

    SiteLattice(const SiteLattice &other)
    {
        alloc();
        memcpy(mpSites, other.mpSites, slice_size(my_rank)*N*N*sizeof(Site));
    }

    SiteLattice &operator=(const SiteLattice &other)
    {
        free();
        alloc();
        memcpy(mpSites, other.mpSites, slice_size(my_rank)*N*N*sizeof(Site));
        return *this;
    }

#if NUM_PROCESSES != 1
    void send(MpiRank rankto)
    {
        if (rankto == my_rank.next()) {
            MPI_Isend(getSlice(endpoint(my_rank) - 1), N*N*sizeof(Site), MPI_BYTE, rankto, my_rank, 
		      MPI_COMM_WORLD, &mRightSend);
        }
        if (rankto.next() == my_rank) {
	    MPI_Isend(getSlice(startpoint(my_rank)), N*N*sizeof(Site), MPI_BYTE, rankto, my_rank + tag_sign_spacer, 
		      MPI_COMM_WORLD, &mLeftSend);
        }
        
    }

    void receive(MpiRank rankfrom)
    {
        if (rankfrom.next() == my_rank) {
            MPI_Irecv(getSlice(endpoint(rankfrom) - 1), N*N*sizeof(Site), MPI_BYTE, rankfrom, rankfrom,
		      MPI_COMM_WORLD, &mRightReceive);
        }
        if (rankfrom == my_rank.next()) {
            MPI_Irecv(getSlice(startpoint(rankfrom)), N*N*sizeof(Site), MPI_BYTE, rankfrom, rankfrom + tag_sign_spacer,
		      MPI_COMM_WORLD, &mLeftReceive);
        }
    }
    void wait()
    {
      MPI_Wait(&mLeftSend,MPI_STATUS_IGNORE);
      MPI_Wait(&mLeftReceive,MPI_STATUS_IGNORE);
      MPI_Wait(&mRightReceive,MPI_STATUS_IGNORE);
      MPI_Wait(&mRightSend,MPI_STATUS_IGNORE);
    }
#endif

    static void rotateData(SiteLattice &sl1, SiteLattice &sl2, SiteLattice &sl3)
    {
        Site *temp = sl1.mpSites;
        sl1.mpSites = sl2.mpSites;
        sl2.mpSites = sl3.mpSites;
        sl3.mpSites = temp;
    }

 protected:
    void alloc()
    {
        mpSites = new Site[slice_size(my_rank)*N*N];
    }

    void free()
    {
        delete [] mpSites;
        mpSites = NULL;
    }

    Site *getSlice(uint xindex)
    {
        return mpSites + N*N*getSliceIndex(xindex);
    }

    const Site *getSlice(uint xindex) const
    {
        return mpSites + N*N*getSliceIndex(xindex);
    }

    Site *mpSites;
#if NUM_PROCESSES != 1
    MPI_Request mLeftSend, mRightSend, mLeftReceive, mRightReceive;
#endif
};


// Utility communication routines
#if NUM_PROCESSES != 1
void send(MpiRank torank, SiteLattice &phi, LinkLattice &A)
{
    phi.send(torank);
    A.send(torank);
}

void receive(MpiRank fromrank, SiteLattice &phi, LinkLattice &A)
{
    phi.receive(fromrank);
    A.receive(fromrank);
}

void sendReceiveRight(SiteLattice &phi, LinkLattice &A)
{
    send(my_rank.next(), phi, A);
    receive(my_rank.prev(), phi, A);
}

void sendReceiveLeft(SiteLattice &phi, LinkLattice &A)
{
    send(my_rank.prev(), phi, A);
    receive(my_rank.next(), phi, A);
}

void wait(SiteLattice &phi, LinkLattice &A)
{
    phi.wait();
    A.wait();
}

void sendReceiveWait(SiteLattice &phi, LinkLattice &A)
{
    sendReceiveLeft(phi, A);
    sendReceiveRight(phi, A);
    wait(phi, A);
}
#endif

// Initialization code, using spherical ansatz form

int epsilon(int i, int j, int k)
{
    weak_assert(i>=0 && i<3);
    weak_assert(j>=0 && j<3);
    weak_assert(k>=0 && k<3);

    if (i == j || j == k || i == k) return 0;

    if (i==0 && j == 1 && k == 2) return 1;
    if (i==1 && j == 2 && k == 0) return 1;
    if (i==2 && j == 0 && k == 1) return 1;

    if (i==0 && j == 2 && k == 1) return -1;
    if (i==2 && j == 1 && k == 0) return -1;
    if (i==1 && j == 0 && k == 2) return -1;

    weak_assert(0);
    return 0;
}

int kroneckerDelta(int i, int j)
{
    weak_assert(i>=0 && i<3);
    weak_assert(j>=0 && j<3);

    return (i == j ? 1 : 0);
}

inline
FloatType sech(FloatType x)
{
    return FloatType(1.0)/cosh(x);
}

FloatType A(FloatType r)
{
    return eps*(0.117*eps + 0.016*eps*r)*pow(sech(eps*r*FloatType(2.0)), FloatType(0.125));
}

FloatType Alpha(FloatType r)
{
    return eps*0.117*eps*r*sech(eps*r/8.0);
}

FloatType Gamma(FloatType)
{
    return FloatType(0.0);
}


FloatType initChi(FloatType rvec[], FloatType r, Direction d, SigmaDirection sd)
{
    FloatType epsterm = FloatType(0.0);
    for (Direction dd = x; dd <= z; dd = Direction(dd+1))
    {
        epsterm += epsilon(d,sd,dd)*rvec[dd]/r;
    }
    return (A(r)*rvec[sd]*rvec[d]/r + Alpha(r)*(kroneckerDelta(d,sd) - rvec[sd]*rvec[d]/(r*r)) + Gamma(r)*epsterm)/(r*g);
}

FloatType Nu(FloatType r)
{
    return (v*g/sqrt(2.0))*eps*(0.017*eps*r*sech(eps*r/5.0));
}

FloatType Mu(FloatType r)
{
    return (v*g/sqrt(2.0))*(1.0 - eps*0.138*sech(eps*r/6.75));
}

Complex Phi(FloatType r)
{
    return Mu(r) + Complex::i*Nu(r);
}

Complex Chi(FloatType r)
{
    return Alpha(r) + Complex::i*(Gamma(r)-FloatType(1.0));
}


Su2Fundamental initPhi(FloatType rvec[], FloatType r)
{
    if (r == FloatType(0.0)) {
        return Su2Fundamental(FloatType(0.0), Mu(r)/(g/sqrt(FloatType(2.0))));
    } else {
        return Su2Fundamental(-Nu(r)*(Complex::i*rvec[x]/r + rvec[y]/r),
                              Mu(r) + Nu(r)*Complex::i*rvec[z]/r)/(g/sqrt(2.0));
    }
}

// Set up initial conditions
void init(SiteLattice &phi0, SiteLattice &phi1, LinkLattice &A0, LinkLattice &A1)
{
    for (uint i = startpoint(my_rank); i < endpoint(my_rank); i++)
    {
        for (uint j = 0; j < N; j++)
        {
            for (uint k = 0; k < N; k++)
            {
                SiteIndex si(i,j,k);

                FloatType rvec[3];
                rvec[x] = delta*(FloatType(i) - FloatType(N)/FloatType(2.0));
                rvec[y] = delta*(FloatType(j) - FloatType(N)/FloatType(2.0));
                rvec[z] = delta*(FloatType(k) - FloatType(N)/FloatType(2.0));

                FloatType rsite = sqrt(rvec[x]*rvec[x] + rvec[y]*rvec[y] + rvec[z]*rvec[z]);

                phi0(si) = initPhi(rvec, rsite);
                phi1(si) = phi0(si);

                for (Direction d = x; d <= z; d = Direction(d+1))
                {
                    rvec[d] += delta/2;
                    FloatType rlink = sqrt(rvec[x]*rvec[x] + rvec[y]*rvec[y] + rvec[z]*rvec[z]);

                    LinkIndex li(si, d);
                    A0(li) = Link(Su2Adjoint(initChi(rvec, rlink, d, sx),
                                             initChi(rvec, rlink, d, sy),
                                             initChi(rvec, rlink, d, sz)), U1());
                    A1(li) = A0(li);
                    rvec[d] -= delta/2;
                }
            }
        }
    }

#if NUM_PROCESSES != 1
    sendReceiveWait(phi0, A0);
    sendReceiveWait(phi1, A1);
#endif
    
}

#ifdef DUMP_DATA
// Generate output for one slice, no-op otherwise
// Allows us to print things only once with parallel processes
class MaybeOstream
{
 public:
    MaybeOstream(const char *filename)
    {
        if (hasDumpSlice(my_rank))
        {
            mpStream = new ofstream(filename);
        }
        else
        {
            mpStream = NULL;
        }
    }

    ~MaybeOstream()
    {
        if (mpStream) {
            delete mpStream;
            mpStream = NULL;
        }
    }

    template<class TYPE>
    MaybeOstream &operator<<(TYPE t)
    {
        if (mpStream)
        {
            *mpStream << t;
        }
        return *this;
    }

    MaybeOstream& operator<< (MaybeOstream& ( *pf )(MaybeOstream&))
    {
        return (*pf)(*this);
    }

    MaybeOstream& maybeEndl()
    {
        if (mpStream)
        {
            *mpStream << endl;
        }
        return *this;
    }
 protected:
    ostream *mpStream;
};

MaybeOstream &endl(MaybeOstream &mos)
{
    mos.maybeEndl();
    return mos;
}
#endif

// Compute energy for our part of the lattice.  Also returns Ebox
FloatType compMyEnergy(const SiteLattice &phi0, const SiteLattice &phi1, const SiteLattice &phi2, const LinkLattice &A0, const LinkLattice &A1, const LinkLattice &A2, FloatType &Ebox)
{
    FloatType E = FloatType(0.0);
    Ebox = FloatType(0.0);

#ifdef DUMP_DATA
    MaybeOstream ofs("graphE.m");
    ofs << " Eden = [";
#endif

    for (uint i = startpoint(my_rank); i < endpoint(my_rank); i++)
    {
        for (uint j = 0; j < N; j++)
        {
            for (uint k = 0; k < N; k++)
            {
                SiteIndex si(i,j,k);
                Su2Fundamental phidot = (phi2(si)-phi0(si))/(2*h);
                Su2Fundamental phiprx = (A1(si,x)*phi1(si+x) - phi1(si))/delta;
                Su2Fundamental phipry = (A1(si,y)*phi1(si+y) - phi1(si))/delta;
                Su2Fundamental phiprz = (A1(si,z)*phi1(si+z) - phi1(si))/delta;
                FloatType potterm = phi1(si).normSquared() - v*v;
                FloatType Eden = phidot.normSquared() + phiprx.normSquared() + phipry.normSquared() + phiprz.normSquared() + lambda*potterm*potterm;

                for (Direction d = x; d <= z; d = Direction(d+1)) {
                    LinkIndex li(si, d);
                    Link Adot = (A2(li)-A0(li))/(2*h);
                    Eden += FloatType(0.5)*Adot.lenSquared();

                    for (Direction dd = Direction(d+1); dd <= z; dd = Direction(dd+1))
                    {
                        Link f1 = A1(si+dd,d)*A1(si+d,dd).conjugate()*A1(si,d).conjugate()*A1(si,dd);
                        Eden += (FloatType(0.5)/(delta*delta))*f1.lenSquared();
                    }
                }
                // Scale to true weak coupling constant
                const FloatType scaling = sqr(g/0.624);
                Eden *= scaling;

#ifdef DUMP_DATA
                if (i==DUMP_SLICE)
                {
                    ofs << Eden << " ";
                }
#endif
                if (delta*sqrt(FloatType((i-N/2)*(i-N/2) + (j-N/2)*(j-N/2) + (k-N/2)*(k-N/2))) <= rbox)
                {
                    Ebox += Eden*delta*delta*delta;
                }
                E += Eden*delta*delta*delta;
                
            }
#ifdef DUMP_DATA
            if (i==DUMP_SLICE)
            {
                ofs << "; ";
            }
#endif
        }
    }

#ifdef DUMP_DATA
    ofs << " ];";
    ofs << endl;
    ofs << "figure(3)" << endl;
    ofs << "contourf(Eden)" << endl;
    ofs << "colorbar" << endl;
#endif
    return E;
}

// Utility output routines
#ifdef DUMP_DATA
void dumpPhiComp(ofstream &os, const SiteLattice &phi, UpDown ud)
{
    os << "[ ";
    for (uint j = 0; j < N; j++)
    {
        for (uint k = 0; k < N; k++)
        {
            SiteIndex si(DUMP_SLICE, j, k);
            os << phi(si)[ud] << " ";
        }
        os << "; ";
    }
    os << "] ;" << endl;
}

void dumpPhi(ofstream &os, const SiteLattice &phi, int index)
{
    os << "phi" << index << "up = ";
    dumpPhiComp(os, phi, up);

    os << "phi" << index << "down = ";
    dumpPhiComp(os, phi, down);
}

ostream &operator<<(ostream &os, Direction d)
{
    switch(d)
    {
     case sx:
        os << "x"; return os;
     case sy:
        os << "y"; return os;
     case sz:
        os << "z"; return os;
     default:
        weak_assert(0);
        return os;
    }
}

ostream &operator<<(ostream &os, SigmaDirection sd)
{
    switch(sd)
    {
     case sx:
        os << "sx"; return os;
     case sy:
        os << "sy"; return os;
     case sz:
        os << "sz"; return os;
     default:
        weak_assert(0);
        return os;
    }
}


void dumpAComp(ofstream &os, const LinkLattice &A, Direction d, int index)
{
    os << "A" << index << d << " = ";
    os << "[ ";
    for (uint j = 0; j < N; j++)
    {
        for (uint k = 0; k < N; k++)
        {
            LinkIndex li(SiteIndex(DUMP_SLICE, j, k), d);
            os << A(li).getU1() << " ";
        }
        os << "; ";
    }
    os << "] ;" << endl;
}

void dumpAComp(ofstream &os, const LinkLattice &A, Direction d, SigmaDirection sd, int index)
{
    os << "A" << index << d << sd << " = ";
    os << "[ ";
    for (uint j = 0; j < N; j++)
    {
        for (uint k = 0; k < N; k++)
        {
            LinkIndex li(SiteIndex(DUMP_SLICE, j, k), d);
            os << A(li).getSu2()[sd] << " ";
        }
        os << "; ";
    }
    os << "] ;" << endl;
}


void dumpA(ofstream &os, const LinkLattice &A, int index)
{

    for (Direction d = x; d <= z; d = Direction(d+1))
    {
        dumpAComp(os, A, d, index);
        for (SigmaDirection sd = sx; sd <= sz; sd = SigmaDirection(sd+1))
        {
            dumpAComp(os, A, d, sd, index);
        }
    }
}


void outputGraphCode(ofstream &ofs, const char *index, const char *comp)
{
    ofs << "subplot(" << index << "),contourf(" << comp << ")" << endl;
    ofs << "title('" << comp << "')" << endl;
    ofs << "colorbar" << endl;
}

// Dump Matlab (C) graphing code
void dumpData(SiteLattice &phi0, SiteLattice &phi1, LinkLattice &A0, LinkLattice &A1)
{
    ofstream ofs("graph.m");
    dumpPhi(ofs, phi0, 0);
    dumpPhi(ofs, phi1, 1);

    dumpA(ofs, A0, 0);
    dumpA(ofs, A1, 1);

    ofs << "figure(1)" << endl;

    outputGraphCode(ofs, "2,2,1", "real(phi1up)");
    outputGraphCode(ofs, "2,2,2", "real(phi1down)");
    outputGraphCode(ofs, "2,2,3", "imag(phi1up)");
    outputGraphCode(ofs, "2,2,4", "imag(phi1down)");

    ofs << "figure(2)" << endl;

    outputGraphCode(ofs, "4,3,1", "A1xsx");
    outputGraphCode(ofs, "4,3,2", "A1xsy");
    outputGraphCode(ofs, "4,3,3", "A1xsz");
    outputGraphCode(ofs, "4,3,4", "A1ysx");
    outputGraphCode(ofs, "4,3,5", "A1ysy");
    outputGraphCode(ofs, "4,3,6", "A1ysz");
    outputGraphCode(ofs, "4,3,7", "A1zsx");
    outputGraphCode(ofs, "4,3,8", "A1zsy");
    outputGraphCode(ofs, "4,3,9", "A1zsz");
    outputGraphCode(ofs, "4,3,10", "A1x");
    outputGraphCode(ofs, "4,3,11", "A1y");
    outputGraphCode(ofs, "4,3,12", "A1z");
}

#endif

// Calculate current or charge using group algebra
inline
LinkLinear calcCurrent(const Su2Fundamental &phi1, const Su2Fundamental &phi2)
{
    return LinkLinear((-g)*Su2AdjointLinear((phi1[up].conjugate()*phi2[down] + phi1[down].conjugate()*phi2[up]).imag_part(),
                                            (phi1[down].conjugate()*phi2[up] - phi1[up].conjugate()*phi2[down]).real_part(),
                                            (phi1[up].conjugate()*phi2[up] - phi1[down].conjugate()*phi2[down]).imag_part()),
                      (-gp)*U1Linear((phi1[up].conjugate()*phi2[up] + phi1[down].conjugate()*phi2[down]).imag_part()));     
}

// Compute Gauss's Law violation for our part of the lattice
FloatType compMyGauss(const SiteLattice &phi1, const SiteLattice &phi2, const LinkLattice &A1, const LinkLattice &A2)
{
#ifdef DUMP_DATA
    MaybeOstream ofs("graphG.m");

    ofs << " Gden = [";
#endif
    FloatType gauss = FloatType(0.0);
    for (uint i = startpoint(my_rank); i < endpoint(my_rank); i++)
    {
        for (uint j = 0; j < N; j++)
        {
            for (uint k = 0; k < N; k++)
            {
                SiteIndex si(i,j,k);
                Su2Fundamental phidot = (phi2(si)-phi1(si))/(h);
                LinkLinear charge = calcCurrent(phidot, phi1(si));
                for (Direction d = x; d <= z; d = Direction(d+1))
                {
		    Link Adotup = A2(si,d)*A1(si,d).conjugate();
                    Link Adotdown = A2(si-d,d).conjugate()*A1(si-d,d);
                    charge += (Adotup.tolinear() + Adotdown.tolinear())/(h*delta);
                }

                FloatType gaussden = charge.len();
                gauss = gauss + delta*delta*delta*gaussden;
#ifdef DUMP_DATA
                if (i==DUMP_SLICE)
                {
                    ofs << gaussden << " ";
                }
#endif
            }
#ifdef DUMP_DATA
            if (i==DUMP_SLICE)
            {
                ofs << "; ";
            }
#endif
        }
    }

#ifdef DUMP_DATA
    ofs << " ];" << endl;
    ofs << "figure(4)" << endl;
    ofs << "contourf(Gden)" << endl;
    ofs << "colorbar" << endl;
#endif

    return gauss;
}


inline uint startpoint_thread(uint threadnum) { return (threadnum*N)/NUM_THREADS; }
inline uint endpoint_thread(uint threadnum)  { return ((threadnum+1)*N)/NUM_THREADS; }

// Central routine:  compute new fields (phi2 and A2) based on fields at the
// two previous time slices (A0, A1, phi0, and phi1).
void nextStep(uint i, ArgStruct *arg)
{
    const SiteLattice &phi0 = *(arg->phi0);
    const LinkLattice &A0 = *(arg->A0);
    const SiteLattice &phi1 = *(arg->phi1);
    const LinkLattice &A1 = *(arg->A1);
    SiteLattice &phi2 = *(arg->phi2);
    LinkLattice &A2 = *(arg->A2);

    // The range in y is divided over threads
    for (uint j = startpoint_thread(arg->threadnum); j < endpoint_thread(arg->threadnum); j++)
    {
        for (uint k = 0; k < N; k++)
        {
            SiteIndex si(i,j,k);
            Su2Fundamental lapphi;
            for (Direction d = x; d <= z; d = Direction(d+1))
            {
                LinkIndex li(si,d);
                Su2Fundamental phiup = A1(li)*phi1(si+d);
                lapphi += phiup + A1(si-d, d).conjugate()*phi1(si-d);
                LinkLinear Adiff = delta*calcCurrent(phi1(si), phiup);
                for (Direction dd = x; dd <= z; dd = Direction(dd+1))
                {
                    if (dd != d)
                    {
                        Link f1 = A1(li)*A1(si+d,dd)*A1(si+dd,d).conjugate()*A1(si,dd).conjugate();
                        Link f2 = A1(li)*A1(si+d-dd,dd).conjugate()*A1(si-dd,d).conjugate()*A1(si-dd,dd);
                        Adiff -= f2.tolinear() + f1.tolinear();
                    }
                }
                A2(li) = Link((A1(li)*A0(li).conjugate()).tolinear() + (h*h/(delta*delta))*Adiff)*A1(li);
            }

            Su2Fundamental phidiff = lapphi + (2*lambda*delta*delta*(v*v - phi1(si).normSquared()) - FloatType(6.0))*phi1(si);
            phi2(si) = FloatType(2.0)*phi1(si)-phi0(si)+(h*h/(delta*delta))*phidiff;
        }
    }
    
}

//  Manage thead communication and call nextStep to do actual computation
void *nextStepThread(void *argsvoid)
{
    ArgStruct *args = (ArgStruct *) argsvoid;

    // The first and last slices need to be passed to neigboring ranks.
    // Do those first so they can go down the pipeline while the rest
    // of the calculation is taking place.

    // First slice
    nextStep(startpoint(my_rank), args);

    // Keep track of how many threads have finished the left-hand slice
#if NUM_PROCESSES != 1
#if NUM_THREADS != 1
    pthread_mutex_lock(&ArgStruct::synchMutex);
    ArgStruct::leftDoneCount++;
    pthread_mutex_unlock(&ArgStruct::synchMutex);
#else
    sendReceiveLeft(*(args->phi2), *(args->A2));
#endif
#endif

    // Last slice
    if (endpoint(my_rank) > startpoint(my_rank) + 1)
    {
        nextStep(endpoint(my_rank) - 1, args);
    }

    // Keep track of how many threads have finished the right-hand slice
#if NUM_PROCESSES != 1
#if NUM_THREADS != 1
    pthread_mutex_lock(&ArgStruct::synchMutex);
    ArgStruct::rightDoneCount++;
    pthread_mutex_unlock(&ArgStruct::synchMutex);
#else
    sendReceiveRight(*(args->phi2), *(args->A2));
#endif
#endif

    // The rest of the slices
    for (uint i = startpoint(my_rank) + 1; i < endpoint(my_rank) - 1; i++)
    {
        nextStep(i, args);
    }

    return NULL;
}

//  Manage MPI communication, call nextStepThread to do computation
void threadCallNextStep(ArgStruct *args)
{
#if NUM_THREADS != 1
    uint n;

#if NUM_PROCESSES != 1
    ArgStruct::leftDoneCount = 0;
    ArgStruct::rightDoneCount = 0;
    pthread_mutex_init(&ArgStruct::synchMutex, NULL);
#endif

    // Spawn threads to do the actual computation
    // All MPI communication is done in the original thread, however
    pthread_t threads[NUM_THREADS];
    for (n=0; n < NUM_THREADS; n++)
    {
        pthread_create(&threads[n], NULL, nextStepThread, &args[n]);
    }

#if NUM_PROCESSES != 1
    // Wait until everyone's done with the left-hand layer
    while (ArgStruct::leftDoneCount < NUM_THREADS)
    {
        sched_yield();
    }
    sendReceiveLeft(*(args->phi2), *(args->A2));

    // Wait until everyone's done with the right-hand layer
    while (ArgStruct::rightDoneCount < NUM_THREADS)
    {
        sched_yield();
    }
    sendReceiveRight(*(args->phi2), *(args->A2));
#endif

    // Wait until everyone's done
    for (n=0; n < NUM_THREADS; n++)
    {
        pthread_join(threads[n], NULL);
    }

#if NUM_PROCESSES != 1
    pthread_mutex_destroy(&ArgStruct::synchMutex);
#endif

#else
    nextStepThread(args);
#endif

#if NUM_PROCESSES != 1
    // Make sure the data has made it to our neighbors by now
    wait(*(args->phi2), *(args->A2));
#endif
}

// Sum energy and gauss violation over the whole lattice (across processes)
void tallyEnergyGauss(FloatType E, FloatType Ebox, FloatType gauss)
{
#if NUM_PROCESSES != 1
    if (my_rank == CONTROL_RANK)
    {
        FloatType Etot = FloatType(0.0);
        FloatType Eboxtot = FloatType(0.0);
        FloatType gausstot = FloatType(0.0);

        for (int fromrank = 0; fromrank < NUM_PROCESSES; fromrank++)
        {
            FloatType Enew = FloatType(0.0);
            FloatType Eboxnew = FloatType(0.0);
            FloatType gaussnew = FloatType(0.0);
            if (fromrank == CONTROL_RANK) {
                Enew = E;
                Eboxnew = Ebox;
                gaussnew = gauss;
            } else {
                MPI_Recv(&Enew, sizeof(FloatType), MPI_BYTE, fromrank, tally_tag_spacer + fromrank, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                MPI_Recv(&Eboxnew, sizeof(FloatType), MPI_BYTE, fromrank, 2*tally_tag_spacer + fromrank, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                MPI_Recv(&gaussnew, sizeof(FloatType), MPI_BYTE, fromrank, 3*tally_tag_spacer + fromrank, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            }
            Etot += Enew;
            Eboxtot += Eboxnew;
            gausstot += gaussnew;
        }
        cout << Etot << " " << gausstot << " " << Eboxtot << endl;
        ofstream curr("currentData.txt");
        curr << Etot << " " << gausstot << " " << Eboxtot << endl;
    } else {
        MPI_Send(&E, sizeof(FloatType), MPI_BYTE, CONTROL_RANK, tally_tag_spacer + my_rank, MPI_COMM_WORLD);
        MPI_Send(&Ebox, sizeof(FloatType), MPI_BYTE, CONTROL_RANK, 2*tally_tag_spacer + my_rank, MPI_COMM_WORLD);
        MPI_Send(&gauss, sizeof(FloatType), MPI_BYTE, CONTROL_RANK, 3*tally_tag_spacer + my_rank, MPI_COMM_WORLD);
    }
#else
    cout << E << " " << gauss << " " << Ebox << endl;
#endif
}

// Unmask floating point exceptions so we see any problems immediately
void initializeFP()
{
#ifdef _MSC_VER
    int fp1 = ~(_EM_ZERODIVIDE | _EM_DENORMAL | _EM_INVALID);
    int fp2 = _MCW_EM;
    _controlfp(fp1, fp2);
#endif
#ifdef __BORLANDC__
    _control87(MCW_EM,MCW_EM);
#endif
#ifdef __GNUC__
    // exceptions for invalid, div by zero and overflow, other bits default.
    // volatile needed by some gcc versions.
    volatile short cw = 0x1372;
    __asm__ volatile("fldcw %0" : : "m" (cw)); /* set control word */
#endif
}

int main(int argc, char **argv)
{
    initializeFP();

#if NUM_PROCESSES != 1
    int totnproc;
    int rank;
    MPI_Init(&argc,&argv);
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Comm_size(MPI_COMM_WORLD,&totnproc);
    my_rank = MpiRank(rank);
#else
    int totnproc = 1;
#endif
    if (NUM_PROCESSES > N || NUM_PROCESSES != totnproc)
    {
        cerr << "Invalid number of processors" << endl;
        exit(-1);
    }
    
#ifdef DUMP_DATA
    MaybeOstream phis("phis.txt");
    MaybeOstream ws("ws.txt");
#endif

    // We will store three time slices:  the third is computed based on the
    // previous two (and then we rotate data to move to the next step).
    SiteLattice phi0;
    SiteLattice phi1;
    SiteLattice phi2;

    LinkLattice A0;
    LinkLattice A1;
    LinkLattice A2;

    // initializes first two time slices
    init(phi0, phi1, A0, A1);
    uint step = 0;
    const uint dumpInterval = 200;
    const FloatType Tmax = FloatType(100.0);

    ArgStruct args[NUM_THREADS];
    setupArgs(args, phi0, phi1, phi2, A0, A1, A2);

    while ((step*h) <= Tmax)
    {
        // returns phi2 and A2
        threadCallNextStep(args);

        // compute energy and Gauss diagonstics
        if ((step % dumpInterval) == 0)
        {
            FloatType Ebox;
            // also returns Ebox
            FloatType E = compMyEnergy(phi0, phi1, phi2, A0, A1, A2, Ebox);
            FloatType gauss = compMyGauss(phi1, phi2, A1, A2);
            tallyEnergyGauss(E, Ebox, gauss);

#ifdef DUMP_DATA
            if (hasDumpSlice(my_rank)) {
                dumpData(phi0, phi1, A0, A1);
            }
#endif
        }
#ifdef DUMP_DATA
        if (hasDumpSlice(my_rank)) {
            SiteIndex si(DUMP_SLICE,DUMP_SLICE,DUMP_SLICE);
            phis << step*h << " " << phi0(si).norm() << " " << phi0(si) << endl;
            ws << step*h << " " << A0(si, x) << " " << A0(si, y) << " " << A0(si, z) << endl;
        }
#endif

        // Transfer data from slice 1 to slice 0, slice 2 to slice 1
        SiteLattice::rotateData(phi0, phi1, phi2);
        LinkLattice::rotateData(A0, A1, A2);

        step++;
    }
#if NUM_PROCESSES != 1
    MPI_Finalize();
#endif

    return 0;
}
